/* syd.c -- System Dynamic Display (V2)

 Copyright (c) 1991, Claus Kalle

 You may distribute under the terms of the GNU General Public
 License as specified in the COPYING file that comes with the syd 2.00
 distribution.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 General Public License for more details.

 Claus Kalle
 Universitaet zu Koeln, Regionales Rechenzentrum
 Robert-Koch-Str. 10
 5000 Koeln 41
 Tel. +49 221 478 5580, Fax. +49 221 478 5590, Mail a0061@rrz.Uni-Koeln.DE

   S Y D
   written by Claus Kalle, University of Cologne Regional Computing Center
   March 1991 for NEC SX3 
   April 1991 by Axel Clauberg for AIX/370 

   define
   _DECOSF for ALPHA specific code
   _SX3 for sx specific code
   _AIX370 the same for 370 */

#define  VERSION "3.1"  /* for DEC OSF/1 Support */

#define BARCHARTMAX 51
#define BARCOL1	22
#define SYD_NLIST_FILE "/tmp/syd_nlist.data"
#ifdef _SX3
#define  SX3CYCLE 0.0000029  /* SX3 cycle time in milliseconds */
#define _SYD_SYSTEM "NEC SX-3"
#define CPUSTATES 5
#define ZEROLLONG 0LL
#endif
#ifdef _AIX370
#define _SYD_SYSTEM "AIX/370"
#define ZEROLLONG 0L
#define CPUSTATES 4
#define p_tmslice p_vecflag
#define p_usertime p_smbeg
#define p_systime p_smend
typedef long mtime_t;
typedef long llong;
typedef unsigned long ullong;
#define XA370
#endif
#ifdef _DECOSF
#define _SYD_SYSTEM "DEC OSF/1"
#define SHOW_UTT
#define CPUSTATES 4
#define ZEROLLONG 0L
typedef long mtime_t;
typedef long llong;
typedef unsigned long ullong;
#define p_tmslice p_stopsig     /* place to save a thingi */
#define p_uid p_ruid
#endif

#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <curses.h>
#include <signal.h>
#include <nlist.h>
#include <malloc.h>
#include <ctype.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#ifndef _DECOSF
#include <sys/var.h>
#endif
#include <sys/sysinfo.h>
#include <sys/utsname.h>
#include <sys/mount.h> /* wegen user.h */
#ifdef _SX3
#include <sys/immu.h>
#include <sys/region.h> /* wegen proc.h */
#include <sys/fs/s5dir.h>
#endif
#include <sys/proc.h>
#include <sys/user.h>
#ifdef _DECOSF
#include <termios.h>
#include <mach.h>
#include <sys/table.h>
#else
#include <sys/trap.h>
#include <sys/reg.h> */
#endif
#ifndef _SX3
typedef struct proc proc_t;
typedef struct user user_t;
#endif  
#ifdef _AIX370
#define SRWAIT SWAIT
#endif
#ifndef _SX3
struct syd_uinfo {
#ifndef _DECOSF
  char comm[ODIRSIZ]; 	/* copy of u.u_comm */
  mtime_t utime;	/* u.u_ru.ru_utime [tv_sec + tv_usec/1000] */ 
  mtime_t stime;	/* u.u_ru.ru_stime  */
  long size;            /* sum of u_tsize, u_dsize and u_ssize */
#else /* _DECOSF */
  time_value_t utime;
  time_value_t stime;
  vm_size_t    vsize;
  vm_size_t    size;
  int          p_pri;
  int          p_basepri;
  long         sleep_time;
  char         statc;
#endif
} *syd_res;
#endif

struct nlist nli[] = {
#ifdef _SX3
#define _X_V 0
	"v",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_SYSINFO 1
	"sysinfo",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_UTSNAME 2
	"utsname",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_FREEMEM1M 3
	"freemem1m",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_FREE32MEM 4
	"free32mem",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_FREEMEM32 5
	"freemem32",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_PROC 6
	"proc",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_BOOTTIME 7
	"boottime",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#endif
#ifdef _AIX370
#define _X_V 0
	"_v",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_SYSINFO 1
	"_sysinfo",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_UTSNAME 2
	"_utsname",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_FREEMEM 3
	"_freemem",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_PROC 4
	"_proc",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_BOOTTIME 5
	"_bootime",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#define _X_MINFO 6
	"_minfo",(llong)0,(short)0,(unsigned short)0,(char)0,(char)0,
#endif
#ifdef _DECOSF
#define _X_NPROC 0
	"_nproc",(llong)0,(short)0,(unsigned short)0,
#define _X_UTSNAME 1
	"_utsname",(llong)0,(short)0,(unsigned short)0,
#define _X_CP_TIME 2
	"_cp_time",(llong)0,(short)0,(unsigned short)0,
#define _X_AVENRUN 3
	"_avenrun",(llong)0,(short)0,(unsigned short)0,
#define _X_PROC 4
	"_proc",(llong)0,(short)0,(unsigned short)0,
#define _X_BOOTTIME 5
	"_boottime",(llong)0,(short)0,(unsigned short)0,
#define _X_VM_SWAP_SPACE 6
	"_vm_swap_space",(llong)0,(short)0,(unsigned short)0,
#define _X_VM_SWAP_IOCOUNT 7
	"_vm_swap_iocount",(llong)0,(short)0,(unsigned short)0,
#define _X_DK_NDRIVE 8
	"_dk_ndrive",(llong)0,(short)0,(unsigned short)0,
#endif
	0,(llong)0,(short)0,(unsigned short)0,(char)0,(char)0
};

extern struct passwd *getpwuid();
#ifdef SIGWINCH
int gwinsize(); /* our SIGWINCH catcher routine */
#endif
int inthnd(); /* our SIGINT catcher routine */
int scralrm(); /* our SIGALRM catcher routine */

char    *usage="usage: %s [-t] [-n sysname] [-s sleep time] [-m mode]\n";
#ifdef _DECOSF
char    *sysname="/vmunix";
#else
char    *sysname="/unix";
#endif
char    *kcoref="/dev/kmem";
char    *coref="/dev/mem";
char    barc[ BARCHARTMAX ];
char	barchar = 'X' ;
char    *prgname;
time_t  curr_time;
struct  tm *curr_tm;
time_t  boottime;
int	memfd,kmemfd;
#ifndef _DECOSF
struct  var var;
struct  sysinfo sysinf[2];
#else
long    cp_time[2][CPUSTATES];
long    avenrun[3];
vm_size_t vmswapspace;
long    pagein[2];
long    pageout[2];
int     swapio[2];
vm_statistics_data_t vmstats;
struct tbl_intr tblintr[2];
struct tbl_ttyinfo tbltty[2];
int    dkndrive;
struct tbl_dkinfo *dkinfo;
int    dixfer[2];
int    dibpc[2];
#endif
#ifdef _AIX370
struct  minfo minf;
int     demandpgs; /* old minfo.demand value */
#endif
int     sysinf_i=0;
struct  utsname utsname;
struct  proc    *proct;
struct  cputm {
  unsigned  pid;
  mtime_t ct;
} *cput;
struct  difft {
  proc_t *prc;
  mtime_t dt;
} *dift=NULL;     /* for saved cpu times of last display */
struct  diffm {
  proc_t *prc;
  int   p_size;
} *difm=NULL;     /* for saved cpu times of last display */
user_t  ubl; /* U-block header (fixed part storage */
#ifdef _SX3
tuser_t tubl; /* task U-block header (fixed part storage */
#endif
#ifdef _DECOSF
struct winsize twinsize;
struct utask utsk;
task_t  task;
task_basic_info_data_t   tbi;
unsigned int tbi_l;
thread_array_t    tharr;
unsigned int tharr_l;
thread_basic_info_t     thi;
thread_basic_info_data_t thidata;
unsigned int thi_l;
#endif
proc_t  *proca; /* kernel proc table address in kmem */
long    maxproc; /* max number of kernel proc entries */
long    lproc; /* length of proc table in use */
long    nproc; /* current number of processes */
long    procstats[9]; /* process status counters */
mtime_t mtotcpu; /* total cpu time found per dsiplay cycle */
int	slptm = 3;
int     done = FALSE;
int     totcpu, totdiff, newcpu[5], oldcpu[5]={0,0,0,0,0}, diff[5];
#ifdef _SX3
int     freemem1m, free32mem, freemem32;
#else
int     freemem;
#endif
long    uptime, updays, uphrs,upmins;
int     termin;
int     dspmod=0; /* 0 normal, 1 percent bars, 2 demo mode, 3 vector perf */
int     vdspsel=0; /* dspmod 3, 0 all procs, 1 only vector procs */
int     ndsp2p=21; /* last line no of processes display during demo mode 2 */

int     cmdcolumns,scrcolumns,scrlines=0,mxplin;

main(argc,argv)
int	argc;
char	**argv;
{
  int i,errflg;
  int c;
  char order;

  extern char *optarg;
  extern int optind;

  prgname=argv[0];
  errflg = 0;
  while((c = getopt(argc,argv,"n:m:s:t")) != -1)
    switch(c) {
    case 'n': sysname = optarg; break;
    case 't': barchar= '#' ; break ;
    case 'm': switch (optarg[0]) {
#ifdef _SX3
      case 'v': dspmod=3; break;
#endif
      case 'd': dspmod=2; break;
      case 'n': dspmod=0; break;
      case 'b': dspmod=1; break;
      default: errflg++; break;
      }
      break;
    case 's': slptm = atoi(optarg); break;
    default: errflg++;
    }
  if (errflg || (optind < argc) ) {fprintf(stderr,usage,prgname); exit(1);}
  tzset();
  for ( i=0; i < BARCHARTMAX; i++ ) barc[ i ]= barchar;
  do_nlist(sysname,nli,sizeof(nli));
  if(nli[0].n_value==0) {fprintf(stderr,"%s: nlist failed\n",prgname);exit(1);}
  if ( (memfd = open(coref,0)) < 0 ) {
    fprintf(stderr,"%s: no access to memory\n",prgname);
    exit(1);
  }
  if ( (kmemfd = open(kcoref,0)) < 0 ) {
    fprintf(stderr,"%s: no access to kernel memory\n",prgname);
    exit(1);
  }
#ifndef _DECOSF
  getkmem(nli[_X_V].n_value,&var,sizeof(struct var),"var");
  getkmem(nli[_X_SYSINFO].n_value,&sysinf[sysinf_i],sizeof(struct sysinfo),"sysinfo");
#else /* DECOSF */
  getkmem(nli[_X_CP_TIME].n_value,&cp_time[sysinf_i][0],CPUSTATES*sizeof(long),"cp_time");
  getkmem(nli[_X_AVENRUN].n_value,avenrun,3*sizeof(long),"avenrun");
  getkmem(nli[_X_VM_SWAP_SPACE].n_value,&vmswapspace,sizeof(vm_size_t),"swapspace");
  getkmem(nli[_X_VM_SWAP_IOCOUNT].n_value,&swapio[sysinf_i],sizeof(int),"swapiocount");
  (void) vm_statistics(task_self(),&vmstats);
  pagein[sysinf_i]=vmstats.pageins; pageout[sysinf_i]=vmstats.pageouts;
  table(TBL_INTR,0,&tblintr[sysinf_i],1,sizeof(struct tbl_intr));
  table(TBL_TTYINFO,0,&tbltty[sysinf_i],1,sizeof(struct tbl_ttyinfo));
  getkmem(nli[_X_DK_NDRIVE].n_value,&dkndrive,sizeof(dkndrive),"dk_ndrive");
  if(dkndrive == 0) {
    fprintf(stderr,"dkndrive = 0\n");
    exit(98);
  }
  if((dkinfo=malloc(dkndrive*sizeof(struct tbl_dkinfo))) == 0) {
    fprintf(stderr,"failed to malloc for dkinfo\n"); exit(1);
  }
  /* table return EINVAL if called directly with dk_ndrive, try to find
     out how many structs are there by trying */
  for(i=0;i<dkndrive;i++) {
    if(table(TBL_DKINFO,i,&dkinfo[i],1,sizeof(struct tbl_dkinfo)) < 0) break;
    if(dkinfo[i].di_name[0] == 'd') break; /* assume end with name = "dk" */
  }
  dkndrive=i; /* reset drives to scan */
  dixfer[sysinf_i]=0;
  dibpc[sysinf_i]=0;
  for(i=0;i<dkndrive;i++) {
    dixfer[sysinf_i] += dkinfo[i].di_xfer;
    dibpc[sysinf_i] += dkinfo[i].di_wds*32/512;
  }
#endif
#ifdef _AIX370
  getkmem(nli[_X_MINFO].n_value,&minf,sizeof(struct minfo),"minfo");
  demandpgs=minf.demand;
#endif
  getkmem(nli[_X_UTSNAME].n_value,&utsname,sizeof(struct utsname),"utsname");
  getkmem(nli[_X_BOOTTIME].n_value,&boottime,sizeof(time_t),"boottime");
#ifdef _SX3
  proca=(proc_t *) nli[_X_PROC].n_value;
#endif
#ifdef _AIX370
  getkmem(nli[_X_PROC].n_value,(char *)&proca,sizeof(struct proc *),"proca");
#endif
#ifdef _DECOSF
  getkmem(nli[_X_PROC].n_value,(char *)&proca,sizeof(struct proc *),"proca");
  getkmem(nli[_X_NPROC].n_value,(char *)&maxproc,sizeof(maxproc),"maxproc");
#else
  maxproc=var.v_proc; 
#endif
  if((proct=(proc_t *)malloc(maxproc*sizeof(struct proc))) == NULL) {
    fprintf(stderr,"%s: cannot malloc for proc table\n",prgname);
    exit(1);
  }
  if((cput=(struct cputm *)malloc((maxproc+1)*sizeof(struct cputm))) == NULL) {
    fprintf(stderr,"%s: cannot malloc for cputm table\n",prgname);
    exit(1);
  }
#ifndef _SX3
  if((syd_res=(struct syd_uinfo *)malloc(maxproc*sizeof(struct syd_uinfo))) == NULL) {
    fprintf(stderr,"%s: cannot malloc for resource table\n",prgname);
    exit(1);
  }
#endif
  cput->pid=0xffff; /* terminator */
  signal(SIGHUP, inthnd);
  signal(SIGINT, inthnd);
  signal(SIGQUIT, inthnd);
  signal(SIGTERM, inthnd);
  if((termin=open("/dev/tty",O_RDWR)) == 0) {
    fprintf(stderr,"%s: open of /dev/tty failed\n",prgname);
    exit(1);
  }
  lincolsetup(); /* this does initscr et al */
  initscr();
  noecho();
  nonl();
  cbreak();
  scrhead();
  scrbody();
  refresh();
#ifdef SIGWINCH
  signal(SIGWINCH,gwinsize);
#endif
  signal(SIGALRM, scralrm);
  alarm(slptm);
  for(;;) {
    if(read(termin,&order,1) == 1) {
      switch (order) {
      case 12:
#ifndef _DECOSF
	redrawwin(stdscr); 
#else
	wrefresh(curscr);
#endif
	touchwin(stdscr); refresh(); /* ^L - refresh screen */ break;
#ifdef _SX3
      case 'v': vdspsel=vdspsel++ & 1;setdspmod(3);/* v - vector mode */ break;
#endif
      case 'b': setdspmod(1); /* b - bar mode */ break;
      case 'n': setdspmod(0); /* n - normal mode */ break;
      case 'd': setdspmod(2); /* d - demo mode */ break;
      case 'e': case 'x': done=1; alarm(0); /* e, x - exit */ break;
      case 'u': setupdate(); /* u - change update cycle */ break;
      case ' ': alarm(0); scralrm(); /* space - update display */ break;
      default: dishelp(); break;
      }
    }
    if(done) break; /* take care of external signals */
  }
  endwin();
  printf("\n");
}

setupdate()
{
  int upd,upi;

  alarm(0); /* end cyclic display */
  upd=0;
  echo();
  while(!upd) {
    attron(A_REVERSE);
    mvprintw(scrlines-1,0,"Seconds between updates ?   ");
    move(scrlines-1,26);
    refresh();
    scanw("%d",&upi);
    if((upi>0) && (upi<60)) upd=upi;
    attroff(A_REVERSE);
  }
  noecho();
  slptm=upd;
  clear();
  scrhead();
  scralrm();
}

dishelp()
{
  alarm(0);
  clear();
  mvprintw(5,15,
              "S Y D  -  System Display Program  Version %s",VERSION);
  mvprintw(7,20,
                   "Type one of the following commands:");
  mvprintw(9,13,
            "n        Normal CPU usage and process activity display");
  mvprintw(10,13,
            "b        Barchart CPU usage and process activity display");
#ifdef _SX3
  mvprintw(11,13,
            "v        Normal CPU usage and process vector usage display");
#endif
  mvprintw(12,13,
            "u        Change display update cycle");
  mvprintw(13,13,
            "<CTRL>L  Refresh screen (e.g. if garbled)");
  mvprintw(14,13,
            "<SPACE>  Update current display immediately");
  mvprintw(15,13,
            "e or x   Exit system display program");
  mvprintw(17,15,
              "Any other input will produce this help screen");
  mvprintw(22,16,
               "Author: Claus Kalle, University of Cologne");
  mvprintw(23,14,
             "Bugs, Suggestions etc. to Kalle@rrz.Uni-Koeln.DE");
  move(scrlines-1,scrcolumns-1);
  refresh();
  clear();
  scrhead(); /* so next command will work properly */
}

setdspmod(val)
int val;
{
  alarm(0); /* kill running display */
  dspmod=val;
  clear();
  scrhead();
  scralrm(); /* start alarms running again */
}

scralrm()  /* alarm routine to display screen */
{
  scrprep();
  scrbody();
  refresh();
  if(!done) {
    signal(SIGALRM, scralrm); 
    alarm(slptm);
  }
}

scrhead()
{
  switch (dspmod) {
  case 0: scrhead0(); break;
  case 1: scrhead1(); break;
  case 2: scrhead1(); break;
#ifdef _SX3
  case 3: scrhead3(); break;
#endif
  }
}

scrhead0()  /* print invariant part of display */
{
  attron(A_REVERSE);
  mvprintw(0,0,"                                                                                ");
  mvprintw(0,16,"-- %s System Information Display Vers. %s --",_SYD_SYSTEM, VERSION);
  mvprintw(0,10,"%s",utsname.release);
  mvprintw(0,1,"%s",utsname.sysname);
  mvprintw(8,0,
#ifdef _DECOSF
" PID  USERNAME NI BP PR S SLP  VSZ  RSS SYSTEMTIME  USERTIME CPU COMMAND        "); /*
 012345678901234567890123456789012345678901234567890123456789012345678901 */
#else
" PID  USERNAME NI BP  JID  P S PRI SIZE SYSTEMTIME  USERTIME CPU COMMAND        "); /*
 012345678901234567890123456789012345678901234567890123456789012345678901 */
#endif
  attroff(A_REVERSE);
  mvprintw(1,0,"CPU Mode Total Current");
#ifndef _DECOSF
  mvprintw(2,0,"Idle"); mvprintw(3,0,"User"); mvprintw(4,0,"Kernel");
  mvprintw(5,0,"Wait"); 
#ifdef _SX3
  mvprintw(6,0,"SXBRK");
#endif
  mvprintw(7,0,"Uptime ");
#else
  mvprintw(2,0,"User"); mvprintw(3,0,"Nice"); mvprintw(4,0,"Kernel");
  mvprintw(5,0,"Idle"); 
  mvprintw(6,0,"Load");
  mvprintw(7,0,"Uptime ");
#endif
}

#ifdef _SX3
scrhead3()  /* print invariant part of display */
{
  mvprintw(0,16,"-- %s System Information Display Vers. %s --",_SYD_SYSTEM, VERSION);
  mvprintw(0,10,"%s",utsname.release);
  mvprintw(0,1,"%s",utsname.sysname);
  mvprintw(1,0,"CPU Mode Total Current");
  mvprintw(2,0,"Idle"); mvprintw(3,0,"User"); mvprintw(4,0,"Kernel");
  mvprintw(5,0,"Wait"); 
  mvprintw(6,0,"SXBRK");
  mvprintw(7,0,"Uptime ");
  mvprintw(8,0,
" PID  USERNAME NI BP  JID  P S PRI SIZE CPU U/S MOPS MFLP VEC VUT COMMAND"); /*
 012345678901234567890123456789012345678901234567890123456789012345678901 */
}
#endif

scrhead1()  /* print invariant part of display */
{
  mvprintw(0,16,"-- %s System Information Display Vers. %s --",_SYD_SYSTEM, VERSION);
  mvprintw(0,10,"%s",utsname.release);
  mvprintw(0,1,"%s",utsname.sysname);
  attron(A_REVERSE);
  mvprintw(2,0,"C P U    U s a g e");
  attroff(A_REVERSE);
  mvprintw(2,BARCOL1,"                       P e r c e n t");
  mvprintw(3,BARCOL1,"0   10   20   30   40   50   60   70   80   90   100");
#ifdef _SX3
  attron(A_ALTCHARSET);
  mvprintw(4,BARCOL1,"lqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqk");
  attroff(A_ALTCHARSET);
#else
  mvprintw(4,BARCOL1,"I----I----I----I----I----I----I----I----I----I----I");
#endif
  attron(A_UNDERLINE);
  mvprintw(4,0,"CPU State");
  attroff(A_UNDERLINE);
#ifndef _DECOSF
  mvprintw(5,0,"Idle"); mvprintw(6,0,"User"); mvprintw(7,0,"Kernel");
  mvprintw(8,0,"Wait"); 
#else
  mvprintw(5,0,"User"); mvprintw(6,0,"Nice"); mvprintw(7,0,"Kernel");
  mvprintw(8,0,"Idle"); 
#endif
#ifdef _SX3
  mvprintw(9,0,"SXBRK");
  attron(A_ALTCHARSET);
  mvprintw(10,BARCOL1,"mqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqj");
  attroff(A_ALTCHARSET);
  mvprintw(11,BARCOL1,"0   10   20   30   40   50   60   70   80   90   100");
#else
  mvprintw(9,BARCOL1,"I----I----I----I----I----I----I----I----I----I----I");
  mvprintw(10,BARCOL1,"0   10   20   30   40   50   60   70   80   90   100");
#endif
  attron(A_REVERSE);
  mvprintw(13,0,"Active Processes  ");
  attroff(A_REVERSE);
  mvprintw(13,BARCOL1,"                       P e r c e n t");
  mvprintw(14,BARCOL1,"0   10   20   30   40   50   60   70   80   90   100");
  attron(A_UNDERLINE);
  mvprintw(15,0,"User"); mvprintw(15,9,"Command");
  attroff(A_UNDERLINE);
#ifdef _SX3
  attron(A_ALTCHARSET);
  mvprintw(15,BARCOL1,"lqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqwqqqqk");
  mvprintw(21,BARCOL1,"mqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqvqqqqj");
  attroff(A_ALTCHARSET);
#else
  mvprintw(15,BARCOL1,"I----I----I----I----I----I----I----I----I----I----I");
  mvprintw(21,BARCOL1,"I----I----I----I----I----I----I----I----I----I----I");
#endif
  mvprintw(22,BARCOL1,"0   10   20   30   40   50   60   70   80   90   100");
}

scrbody()
{
  switch (dspmod) {
  case 0: scrbody0_3(); break;
  case 1: scrbody1(); break;
  case 2: scrbody2(); break;
  case 3: scrbody0_3(); break;
  }
}

scrbody0_3()  /* print display body */
{
  int i;

  time(&curr_time);
  curr_tm=localtime(&curr_time);
  attron(A_REVERSE);
  mvprintw(0,71,"%02d:%02d:%02d",curr_tm->tm_hour,curr_tm->tm_min,curr_tm->tm_sec);
  attroff(A_REVERSE);
  uptime=curr_time-boottime;
  uptime += 30;
  updays = uptime / (60*60*24);
  uptime %= (60*60*24);
  uphrs = uptime / (60*60);
  uptime %= (60*60);
  upmins = uptime / 60;
  if (updays > 0) {
    mvprintw(7,8,"%d day%s, %2d:%02d ",updays, updays>1?"s":"", uphrs, upmins);
  } else
    mvprintw(7,8,"%d hr%s, %d min%s  ", uphrs, uphrs>1?"s":"", upmins, upmins>1?"s":"");
#ifdef _DECOSF
  for(i=0;i<3;i++)
    mvprintw(6,6+i*6,"%2d.%02d",avenrun[i]/1000,(avenrun[i]%1000)/10);
#endif
  totcpu=0; totdiff=0;
  for(i=0;i<CPUSTATES;i++){
#ifndef _DECOSF
    newcpu[i]=sysinf[sysinf_i].cpu[i];
#else
    newcpu[i]=cp_time[sysinf_i][i];
#endif
    diff[i]=newcpu[i]-oldcpu[i];
    totcpu+=newcpu[i]; totdiff+=diff[i];
    oldcpu[i]=newcpu[i];
  }
  totcpu=totcpu/100;
  for(i=0;i<CPUSTATES;i++){
    newcpu[i]=newcpu[i]/totcpu;
    diff[i]=(diff[i]*100)/totdiff;
    mvprintw(i+2,9,"%3d%%    %3d%%",newcpu[i],diff[i]);
  }
  sysinf_i=(sysinf_i+1) & 1;
#ifdef _SX3
  mvprintw(1,39,"bread  %6ld%6ld",sysinf[sysinf_i].bread,sysinf[sysinf_i].stread);
  mvprintw(2,39,"bwrite %6ld%6ld",sysinf[sysinf_i].bwrite,sysinf[sysinf_i].stwrite);
  mvprintw(3,39,"lread  %6ld%6ld",sysinf[sysinf_i].lread,sysinf[sysinf_i].slread);
  mvprintw(4,39,"lwrite %6ld%6ld",sysinf[sysinf_i].lwrite,sysinf[sysinf_i].slwrite);
  mvprintw(5,39,"phread %6ld%6ld",sysinf[sysinf_i].phread,sysinf[sysinf_i].sbread);
  mvprintw(6,39,"phwrite%6ld%6ld",sysinf[sysinf_i].phwrite,sysinf[sysinf_i].sbwrite);
#endif
#ifdef _AIX370
  mvprintw(1,40,"bread  %10ld",sysinf[sysinf_i].bread);
  mvprintw(2,40,"bwrite %10ld",sysinf[sysinf_i].bwrite);
  mvprintw(3,40,"lread  %10ld",sysinf[sysinf_i].lread);
  mvprintw(4,40,"lwrite %10ld",sysinf[sysinf_i].lwrite);
  mvprintw(5,40,"phread %10ld",sysinf[sysinf_i].phread);
  mvprintw(6,40,"phwrite%10ld",sysinf[sysinf_i].phwrite);
#endif
#ifdef _DECOSF
  mvprintw(1,40,"devintr%10ld",tblintr[sysinf_i].in_devintr);
  mvprintw(2,40,"cntxtsw%10ld",tblintr[sysinf_i].in_context);
  mvprintw(3,40,"syscall%10ld",tblintr[sysinf_i].in_syscalls);
  mvprintw(4,40,"forks  %10ld",tblintr[sysinf_i].in_forks);
  mvprintw(5,40,"vforks %10ld",tblintr[sysinf_i].in_vforks);
  mvprintw(6,40,"ttinch %10ld",tbltty[sysinf_i].ti_nin);
  mvprintw(7,40,"ttoutch%10ld",tbltty[sysinf_i].ti_nout);
  mvprintw(1,60,"freeswp%10ld",vmswapspace*vm_page_size/1024);
  mvprintw(2,60,"freemem%10ld",(vmstats.free_count+vmstats.inactive_count)
	   *vm_page_size/1024);
  mvprintw(3,60,"pagein %10ld",pagein[sysinf_i]);
  mvprintw(4,60,"pageout%10ld",pageout[sysinf_i]);
  mvprintw(5,60,"swapio %10ld",swapio[sysinf_i]);
  mvprintw(6,60,"dkxfers%10ld",dixfer[sysinf_i]);
  mvprintw(7,60,"dkblcks%10ld",dibpc[sysinf_i]);
#else
  mvprintw(1,60,"swapin %10ld",sysinf[sysinf_i].swapin);
  mvprintw(2,60,"swapout%10ld",sysinf[sysinf_i].swapout);
#endif
#ifdef _SX3
  mvprintw(3,60,"free 1m%10ld",freemem1m);
  mvprintw(4,60,"freem32%10ld",free32mem);
  mvprintw(5,60,"free 32%10ld",freemem32);
#endif
#ifdef _AIX370
  mvprintw(3,60,"freemem%10ld",freemem);
  mvprintw(4,60,"freeswp%10ld",minf.freeswap);
  mvprintw(5,60,"demandp%10ld",minf.demand-demandpgs);
#endif
#ifndef _DECOSF
  mvprintw(7,40,"syscall%10ld",sysinf[sysinf_i].syscall);
  mvprintw(6,60,"sysfork%10ld",sysinf[sysinf_i].sysfork);
  mvprintw(7,60,"sysexec%10ld",sysinf[sysinf_i].sysexec);
#endif
  doprocs();
  clrtobot();
  mvprintw(1,25,"  Processes");
#ifdef _DECOSF
  mvprintw(2,25,"Runnable %4d",procstats[1]);
  mvprintw(3,25,"Sleeping %4d",procstats[2]);
  mvprintw(4,25,"Inactive %4d",procstats[3]);
  mvprintw(5,25,"UnintWait%4d",procstats[4]);
  mvprintw(6,25,"Stopped  %4d",procstats[5]);
  mvprintw(7,25,"Other    %4d",procstats[6]);
#else
#ifdef SONPROC
  mvprintw(2,25,"Runnable %4d",procstats[SRUN]+procstats[SONPROC]);
#else
  mvprintw(2,25,"Runnable %4d",procstats[SRUN]);
#endif
  mvprintw(3,25,"Sleeping %4d",procstats[SSLEEP]);
  mvprintw(4,25,"MemWait  %4d",procstats[SXBRK]);
  mvprintw(5,25,"Waiting  %4d",procstats[SRWAIT]);
  mvprintw(6,25,"Other    %4d",procstats[SZOMB]+procstats[SSTOP]+procstats[SIDL]);
#endif
  move(scrlines-1,scrcolumns-1);
}

scrbody1()  /* print display body */
{
  int i,j;

  time(&curr_time);
  curr_tm=localtime(&curr_time);
  mvprintw(0,71,"%02d:%02d:%02d",curr_tm->tm_hour,curr_tm->tm_min,curr_tm->tm_sec);
  totdiff=0;
  for(i=0;i<CPUSTATES;i++){
#ifndef _DECOSF
    newcpu[i]=sysinf[sysinf_i].cpu[i];
#else
    newcpu[i]=cp_time[sysinf_i][i];
#endif
    diff[i]=newcpu[i]-oldcpu[i];
    totdiff+=diff[i];
    oldcpu[i]=newcpu[i];
  }
  for(i=0;i<CPUSTATES;i++){
    j=(diff[i]*BARCHARTMAX)/totdiff;
    diff[i]=(diff[i]*100)/totdiff;
    mvprintw(i+5,16,"%3d%%",diff[i]);
    if((j<0) || (j>BARCHARTMAX)) j=0;
    barc[j]='\0';
    attron(A_ALTCHARSET);
    mvprintw(i+5,BARCOL1,"%s",barc);
    attroff(A_ALTCHARSET);
    clrtoeol();
    barc[j]= barchar;
  }
  sysinf_i=(sysinf_i+1) & 1;
  doprocs();
  move(scrlines-1,scrcolumns-1);
}

scrbody2()  /* print display body - demo mode */
{
  int i,j,k;
  char ch;

  alarm(0); /* This is a serailly executed demo code */
  for(;;) {
    clear();
    attron(A_BLINK);
    mvprintw(12,22,     "... Waiting for \"Startup\" Key ...");
    attroff(A_BLINK);
    move(scrlines-1,scrcolumns-1);
    refresh();
    scanw("%c",&ch);
    if(ch == '.') break;
  }
  clear();
  scrhead();
  time(&curr_time);
  curr_tm=localtime(&curr_time);
  mvprintw(0,71,"%02d:%02d:%02d",curr_tm->tm_hour,curr_tm->tm_min,curr_tm->tm_sec);
#ifdef _SX3
  for(i=0;i<5;i++) { /* print wait startup message */
#else
  for(i=0;i<4;i++) { /* print wait startup message */
#endif
    mvprintw(i+5,9,"%3d%%",0);
    mvprintw(i+5,BARCOL1,"%s","- Waiting for System Startup -");
  }
  move(scrlines-1,scrcolumns-1);
  refresh();
  attron(A_BLINK|A_REVERSE);
  mvprintw(1,22,"S Y S T E M    S T A R T I N G    U P");
  attroff(A_BLINK|A_REVERSE);
  sleep(10);
#ifdef _SX3
  for(i=0; i<5; i++){
#else
  for(i=0; i<4; i++){
#endif
    for(j=0;j<101;j+=2) {
      mvprintw(i+5,9,"%3d%%",j);
      k=(j*60)/100;
      barc[k]='\0';
      attron(A_ALTCHARSET);
      mvprintw(i+5,BARCOL1,"%s",barc);
      attroff(A_ALTCHARSET);
      barc[k]= barchar;
      refresh();
    }
    for(j=100;j>=0;j-=2) {
      mvprintw(i+5,9,"%3d%%",j);
      k=(j*60)/100;
      barc[k]='\0';
      attron(A_ALTCHARSET);
      mvprintw(i+5,BARCOL1,"%s",barc);
      attroff(A_ALTCHARSET);
      clrtoeol();
      barc[k]= barchar;
      refresh();
    }
    sleep(1);
  }
  move(scrlines-1,scrcolumns-1);
  refresh();
  move(1,0); clrtoeol();
  slptm=3;
  dspmod=1; /* bar codes */
  sleep(10);
  ndsp2p=17; /* start to display 1 process */
  scralrm();

}

scrprep() /* prepare next display */
{
  int i,j;

#ifdef _AIX370
  demandpgs=minf.demand;
  getkmem(nli[_X_MINFO].n_value,&minf,sizeof(struct minfo),"minfo");
#endif
#ifndef _DECOSF
  getkmem(nli[_X_SYSINFO].n_value,&sysinf[sysinf_i],sizeof(struct sysinfo),"sysinfo");
#else DECOSF */
  getkmem(nli[_X_CP_TIME].n_value,&cp_time[sysinf_i][0],CPUSTATES*sizeof(long),"cp_time");
  getkmem(nli[_X_AVENRUN].n_value,avenrun,3*sizeof(long),"avenrun");
  getkmem(nli[_X_VM_SWAP_SPACE].n_value,&vmswapspace,sizeof(vm_size_t),"swapspace");
  getkmem(nli[_X_VM_SWAP_IOCOUNT].n_value,&swapio[sysinf_i],sizeof(int),"swapiocount");
  (void) vm_statistics(task_self(),&vmstats);
  pagein[sysinf_i]=vmstats.pageins; pageout[sysinf_i]=vmstats.pageouts;
  table(TBL_INTR,0,&tblintr[sysinf_i],1,sizeof(struct tbl_intr));
  table(TBL_TTYINFO,0,&tbltty[sysinf_i],1,sizeof(struct tbl_ttyinfo));
  for(i=0;i<dkndrive;i++)
    table(TBL_DKINFO,i,&dkinfo[i],1,sizeof(struct tbl_dkinfo));
  dixfer[sysinf_i]=0;
  dibpc[sysinf_i]=0;
  for(i=0;i<dkndrive;i++) {
    dixfer[sysinf_i] += dkinfo[i].di_xfer;
    dibpc[sysinf_i] += dkinfo[i].di_wds*32/512;
  }
#endif
  i=(sysinf_i+1) & 1;
#ifndef _DECOSF
#ifdef _SX3
  sysinf[i].stread=sysinf[sysinf_i].stread-sysinf[i].stread;
  sysinf[i].stwrite=sysinf[sysinf_i].stwrite-sysinf[i].stwrite;
  sysinf[i].slread=sysinf[sysinf_i].slread-sysinf[i].slread;
  sysinf[i].slwrite=sysinf[sysinf_i].slwrite-sysinf[i].slwrite;
  sysinf[i].sbread=sysinf[sysinf_i].sbread-sysinf[i].sbread;
  sysinf[i].sbwrite=sysinf[sysinf_i].sbwrite-sysinf[i].sbwrite;
#endif
  sysinf[i].bread=sysinf[sysinf_i].bread-sysinf[i].bread;
  sysinf[i].bwrite=sysinf[sysinf_i].bwrite-sysinf[i].bwrite;
  sysinf[i].lread=sysinf[sysinf_i].lread-sysinf[i].lread;
  sysinf[i].lwrite=sysinf[sysinf_i].lwrite-sysinf[i].lwrite;
  sysinf[i].phread=sysinf[sysinf_i].phread-sysinf[i].phread;
  sysinf[i].phwrite=sysinf[sysinf_i].phwrite-sysinf[i].phwrite;
  sysinf[i].swapin=sysinf[sysinf_i].swapin-sysinf[i].swapin;
  sysinf[i].swapout=sysinf[sysinf_i].swapout-sysinf[i].swapout;
  sysinf[i].syscall=sysinf[sysinf_i].syscall-sysinf[i].syscall;
  sysinf[i].sysfork=sysinf[sysinf_i].sysfork-sysinf[i].sysfork;
  sysinf[i].sysexec=sysinf[sysinf_i].sysexec-sysinf[i].sysexec;
#else /* DECOSF */
  swapio[i]=swapio[sysinf_i]-swapio[i];
  pagein[i]=pagein[sysinf_i]-pagein[i];
  pageout[i]=pageout[sysinf_i]-pageout[i];
  tblintr[i].in_devintr=tblintr[sysinf_i].in_devintr-tblintr[i].in_devintr;
  tblintr[i].in_context=tblintr[sysinf_i].in_context-tblintr[i].in_context;
  tblintr[i].in_syscalls=tblintr[sysinf_i].in_syscalls-tblintr[i].in_syscalls;
  tblintr[i].in_forks=tblintr[sysinf_i].in_forks-tblintr[i].in_forks;
  tblintr[i].in_vforks=tblintr[sysinf_i].in_vforks-tblintr[i].in_vforks;
  tbltty[i].ti_nin=tbltty[sysinf_i].ti_nin-tbltty[i].ti_nin;
  tbltty[i].ti_nout=tbltty[sysinf_i].ti_nout-tbltty[i].ti_nout;
  dixfer[i]=dixfer[sysinf_i]-dixfer[i];
  dibpc[i]=dibpc[sysinf_i]-dibpc[i];
#endif
#ifdef _SX3
  getkmem(nli[_X_FREEMEM1M].n_value,&freemem1m,sizeof(int),"freemem1m");
  getkmem(nli[_X_FREE32MEM].n_value,&free32mem,sizeof(int),"free32mem");
  getkmem(nli[_X_FREEMEM32].n_value,&freemem32,sizeof(int),"freemem32");
#endif
#ifdef _AIX370
  getkmem(nli[_X_FREEMEM].n_value,&freemem,sizeof(int),"freemem");
#endif
}

inthnd()
{
  done=TRUE;
}

lincolsetup()
{
  int i;
#ifndef TIOCGWINSZ
  scrcolumns = tigetnum("cols");
  scrlines = tigetnum("lines");
#else
  ioctl(termin,TIOCGWINSZ,&twinsize);
  scrcolumns = twinsize.ws_col;
  scrlines = twinsize.ws_row;
#endif
  if(scrcolumns < 5) scrcolumns=80;
  if(scrlines < 5) scrlines=24;
  if(scrcolumns < 80) {
    endwin();
    fprintf(stderr,"%s needs at least 80 columns for display\n",prgname);
    exit(1);
  }
  cmdcolumns = scrcolumns - 65 ;
  mxplin=scrlines-9; /* number of lines in process list display */
  if((dift=(struct difft *)realloc(dift,mxplin*sizeof(struct difft))) == NULL) {
    fprintf(stderr,"%s: cannot malloc for dift table\n",prgname);
    exit(1);
  }
  if((difm=(struct diffm *)realloc(difm,mxplin*sizeof(struct diffm))) == NULL) {
    fprintf(stderr,"%s: cannot malloc for difm table\n",prgname);
    exit(1);
  }
}

#ifdef SIGWINCH
gwinsize() /* called at SIGWINCH time, setup all screen stuff again */
{
  alarm(0); /* turn off periodic display */
  endwin();
  lincolsetup();
  initscr(); /* must do this to reassure display size to curses */
  noecho();
  nonl();
  cbreak();
  scrhead();
#ifdef SYD_DEBUG
  mvprintw(0,0,"c=%d,l=%d",scrcolumns,scrlines);
#endif
  scralrm();
}
#endif

getkmem(addr,to,nbytes,varname) /* read kmem at addr into to, nbytes bytes */
long addr, nbytes;
char *to, *varname;
{
  if (lseek(kmemfd,(llong)addr,0) != -1) {
    if (nbytes == read(kmemfd,to,nbytes)) return(1);
  }
  if (varname != 0L) {
    fprintf(stderr,"\n%s: kmem - cannot read %s\n",prgname,varname);
    exit(1);
  } else {
    return(0);
  }
}

getmem(addr,to,nbytes,varname) /* read system memory */
     char  *addr; /* address in /dev/mem */
     long  nbytes;/* number bytes to read */
     char  *to;   /* where to put read data */
     char  *varname; /* name for error message, NULL for nonzero return code */
{
  if (lseek(memfd,addr,0) != -1) {
    if (nbytes == read(memfd,to,nbytes)) return(1);
  }
  if (varname) {
    fprintf(stderr,"\n%s: mem - cannot read %s\n",prgname,varname);
    exit(1);
  } else {
    return(0);
  }
}

char *uid2name(muid)
int muid;
{
  static struct uidnames {
    int uid;
    char user[8];
  } uidn[30];
  static int nuids=0;
  int i;
  struct passwd *pwde;

  if(nuids) {
    for(i=0; i<nuids; i++) {
      if(muid == uidn[i].uid) return (uidn[i].user);
    }
    if(nuids==30) nuids=0; /* rebuild table if full nad not found */
  }
  uidn[nuids].uid=muid;
  if(pwde=getpwuid(muid)) {
    strncpy(uidn[nuids].user,pwde->pw_name,8);
  } else {
    sprintf(uidn[nuids].user,"UID%5d",muid);
  }
  return(uidn[nuids++].user);
}

doprocs()
{
  int i,j,k,kk;
  long l;
  struct cputm *pptr;
  mtime_t tcput, dcput;
  time_value_t tcputv;
  int prsort[5]; /* sort list for bar display */
#ifdef _DECOSF
  getkmem(nli[_X_NPROC].n_value,(char *)&nproc,sizeof(nproc),"nproc");
  lproc=nproc*sizeof(struct proc);
#else
  struct var v;

  getkmem(nli[_X_V].n_value,(char *)&v,sizeof(struct var),"var");
  lproc=(long) ((llong)v.ve_proc - (llong)proca);
  
  nproc=lproc / sizeof(struct proc);
#endif

  getkmem((ulong) proca,(char *)proct,lproc,"proc entries");
  mtotcpu=ZEROLLONG;
  for (i=0;i<(int) 9;i++) procstats[i]=0L;
  for(i=0;i<mxplin;i++) {dift[i].dt=-1;dift[i].prc=NULL;
		     difm[i].p_size=-1;difm[i].prc=NULL;}
  for(i=0;i<nproc;i++) { 
#ifdef _SX3
    if(proct[i].p_task.t_stat<9) procstats[proct[i].p_task.t_stat]++;
#else
#ifndef _DECOSF
    if(proct[i].p_stat<9) procstats[proct[i].p_stat]++;
#endif
#endif
#ifdef _SX3
    if((proct[i].p_task.t_stat > 0) && (proct[i].p_task.t_stat < 9)) {
      tcput=proct[i].p_hutime+proct[i].p_hstime;
#else
#ifndef _DECOSF
    syd_res[i].comm[0]='\0';
    syd_res[i].utime=0;
    syd_res[i].stime=0;
#endif
    syd_res[i].size=0;
#ifdef _DECOSF
    syd_res[i].utime.seconds=0;
    syd_res[i].utime.microseconds=0;
    syd_res[i].stime.seconds=0;
    syd_res[i].stime.microseconds=0;
    syd_res[i].vsize=0;
    syd_res[i].statc=' ';
    if((proct[i].p_stat > 0) && (proct[i].p_stat < 8)) {
      if(task_by_unix_pid(task_self(), proct[i].p_pid, &task) != KERN_SUCCESS){
	syd_res[i].statc='Z';
      } else {
	tbi_l=TASK_BASIC_INFO_COUNT;
	if(task_info(task, TASK_BASIC_INFO, (task_info_t) &tbi, &tbi_l)
	   != KERN_SUCCESS) {
	  syd_res[i].statc='Z';
	} else {
	  int minstat=99,mstato,mcurp=1000,mbasp=1000,mslpt=999;
	  char	mach_state_table[] = " RUSITH?";
	  syd_res[i].size=tbi.resident_size;
	  syd_res[i].vsize=tbi.virtual_size;
	  (void) task_threads(task, &tharr, &tharr_l);
	  syd_res[i].utime=tbi.user_time;
	  /* the next line corrects a BUG: hiword contains sometimes rubbish */
	  tbi.system_time.microseconds &= 0x00000000ffffffff;
	  syd_res[i].stime=tbi.system_time;
	  thi= &thidata;
	  for(j=0; j < tharr_l; j++) {
	    thi_l=THREAD_BASIC_INFO_COUNT;
	    if(thread_info(tharr[j],THREAD_BASIC_INFO,
			   (thread_info_t) thi, &thi_l) == KERN_SUCCESS) {
	      time_value_add(&syd_res[i].utime,&thi->user_time);
	      time_value_add(&syd_res[i].stime,&thi->system_time);
	      if(minstat>thi->run_state) minstat=thi->run_state;
	      if(mcurp>thi->cur_priority) mcurp=thi->cur_priority;
	      if(mbasp>thi->base_priority) mbasp=thi->base_priority;
	      if(mslpt>thi->sleep_time) mslpt=thi->sleep_time;
	    }
	  }
	  switch (minstat) {
	  case TH_STATE_RUNNING:      mstato=1; procstats[1]++; break;
	  case TH_STATE_UNINTERRUPTIBLE: mstato=2; procstats[4]++; break;
	  case TH_STATE_WAITING:      mstato=(thi->sleep_time > 20) ? 4 : 3; procstats[mstato-1]++; break;
	  case TH_STATE_STOPPED:      mstato=5; procstats[5]++; break;
	  case TH_STATE_HALTED:       mstato=6; procstats[6]++; break;
	  default:                    mstato=7; procstats[6]++; break;
	  }
	  syd_res[i].statc=mach_state_table[mstato];
	  syd_res[i].p_pri=mcurp;
	  syd_res[i].p_basepri=mbasp;
	  syd_res[i].sleep_time=mslpt;
	  vm_deallocate(task_self(),tharr,tharr_l);
	}
      }
      tcputv=syd_res[i].utime;
      time_value_add(&tcputv,&syd_res[i].stime);
      tcput=tcputv.seconds*1000 + tcputv.microseconds/1000;
#else
    if((proct[i].p_stat > 0) && (proct[i].p_stat < 8)) {
      user_t u;
      if(proct[i].p_flag & SLOAD) {
	if(getkmem ((long)uaddress(proct[i].p_addr.p_caddr)|0x30000000, 
		    (char *)&u,sizeof(user_t) , (char *) 0)) { /* read u-block */
	  strcpy (syd_res[i].comm,u.u_comm);
	  syd_res[i].utime=(u.u_ru.ru_utime.tv_sec*1000)+(u.u_ru.ru_utime.tv_usec/1000);
	  syd_res[i].stime=(u.u_ru.ru_stime.tv_sec*1000)+(u.u_ru.ru_stime.tv_usec/1000);
	  syd_res[i].size=u.u_tsize+u.u_dsize+u.u_ssize;
	}
      }
      tcput=syd_res[i].utime + syd_res[i].stime;
#endif
#endif
      proct[i].p_tmslice=tcput; /* save for below */
      dcput=tcput;
      for(pptr=cput;pptr->pid != (int)0xffff;pptr++) {
	if(proct[i].p_pid == pptr->pid) {
	  dcput-=pptr->ct;
	  break;
	}
      }
      if(dcput != 0) {
      mtotcpu+=dcput; 
	for(k=0;k<mxplin;k++) {
	  if(dcput>dift[k].dt) {
	    for(j=mxplin-1; j>k;j--) { 
	      dift[j].dt=dift[j-1].dt;
	      dift[j].prc=dift[j-1].prc;
	    }
	    dift[k].dt=dcput;
	    dift[k].prc= &proct[i];
	    break;
	  }
	}
      }
    }
  }
  pptr=cput;
  for(i=0;i<nproc;i++) { /* save cpu time for next cycle */
    pptr->pid=proct[i].p_pid;
    (pptr++)->ct=proct[i].p_tmslice; /* see above */
  }
  pptr->pid=0xffff; /* terminator */
  if(dspmod==0 || dspmod==3) { /* build p_size diff list */
    for(i=0;i<nproc;i++) {
      for(k=0;k<mxplin;k++) { /* build list difm of p_size */
#ifdef _SX3
	if((proct[i].p_stat != 0) && (proct[i].p_size>difm[k].p_size)) {
	  for(j=mxplin-1; j>k;j--) { 
	    difm[j].p_size=difm[j-1].p_size;
	    difm[j].prc=difm[j-1].prc;
	  }
	  difm[k].p_size=proct[i].p_size;
	  difm[k].prc= &proct[i];
	  break;
	}
#else
	if((proct[i].p_stat !=0) &&
	   ((long) syd_res[i].size > (long) difm[k].p_size)) {
	  for(j=mxplin-1; j>k;j--) { 
	    difm[j].p_size=difm[j-1].p_size;
	    difm[j].prc=difm[j-1].prc;
	  }
	  difm[k].p_size=syd_res[i].size;
	  difm[k].prc= &proct[i];
	  break;
	}
#endif
      }
    }
  }
#ifdef SYD_DEBUG
  mvprintw(8,0,"mxplin=%d",mxplin);
#endif
  if(dspmod==0 || dspmod==3) {
    j=9; /* start line for proc display */
    for(i=0;i<mxplin && j< scrlines ;i++) {
      if(dift[i].dt != -1) {
	displproc(dift[i].prc,&j,dift[i].dt); j++;
#ifdef _SX3
	dift[i].prc->p_hstime= (mtime_t) -1; /* flag proc was displayed */
#else
	dift[i].prc->p_tmslice= 0xffffffff; /* flag proc was displayed */
#endif
      } else
	break;
    }
    for(i=0;i<mxplin && j< scrlines ;i++) {
      if(difm[i].p_size != -1) {
#ifdef _SX3
	if(difm[i].prc->p_hstime != (mtime_t) -1) {
#else
	if(difm[i].prc->p_tmslice !=  0xffffffff) {
#endif
	  displproc(difm[i].prc,&j,(mtime_t) 0); j++;
	}
      } else
	break;
    }
  }
  if (dspmod==1) { /* bar mode */
    if(mxplin>1) { /* sort processes to display by pid */
      prsort[0]=0;
      k=0;
      for(i=1;(i<5) && (i<mxplin);i++) {
	if(dift[i].dt != -1) {
	  for(j=0;j<=k;j++) {
	    if(dift[i].prc->p_pid < dift[prsort[j]].prc->p_pid) {
	      for(kk=k;kk>=j;kk--) prsort[kk+1]=prsort[kk];
	      prsort[j]=i;
	      break;
	    }
	  }
	  if(j>k) prsort[j]=i;
	  k++;
	}
      }
    }
    j=16; /* start line for proc display */
    for(i=0;(i<=k) && (j<ndsp2p);i++) 
      displpbar(dift[prsort[i]].prc,j++,dift[prsort[i]].dt);
    while(j<21) { move(j++,0); clrtoeol(); }
    if(ndsp2p==20) {ndsp2p++; slptm=1;}
    if(ndsp2p<20) ndsp2p++;
  }
}

char *lastcmd(prctbl) /* return last command string from U-block */
struct proc *prctbl;
{
#ifdef _SX3
  pde_t  ubpgi; /* U-block page table entry */
  user_t *ubaddr;/* U-block real memory address */

  if(! (prctbl->p_flag & SULOAD)) return("(swapped)");
  getkmem((long)ubptbl(prctbl),&ubpgi,sizeof(pde_t),"u-block pde");
  ubaddr=(user_t *) ((ubpgi.pgi.pg_pde & PG_ADDR) << PG_SHFT32);
  getmem(ubaddr,&ubl,sizeof(user_t),"u-block");
  if(/* (ubl.u_ubsize)  && R1.32 */ (ubl.u_procp-proca == prctbl-proct) )  {
    return(ubl.u_psargs);
  } else {
    return("<defunct>");
  }
#endif
#ifdef _AIX370
  if (((prctbl->p_flag & SSPCL)== SSPCL) || ((prctbl->p_flag & SSYS)== SSYS))
    return ("<system process>");
  else if (! (prctbl->p_flag & SLOAD)) 
    return ("<swapped>");
  else {
    return (syd_res[((long)prctbl-(long)proct)/sizeof(struct proc)].comm); }
#endif
#ifdef _DECOSF
  if (! (prctbl->p_flag & SLOAD)) 
    return ("<swapped>");
  else if (prctbl->utask == 0)
    return ("<idle>");
  else {
    getkmem(prctbl->utask,&utsk,sizeof(struct utask),"utask block");
    return (utsk.uu_comm);
  }
#endif
}

#ifdef _SX3
long tmops(prctbl) /* return mops from task block */
struct proc *prctbl;
{
  pde_t  tbpgi; /* T-block page table entry */
  tuser_t *tbaddr;/* T-block real memory address */

  if(! (prctbl->p_task.t_flag & STULOAD)) return(-1);
  getkmem((long)tbptbl((task_t *) prctbl),&tbpgi,sizeof(pde_t),"tu-block pde");
  tbaddr=(tuser_t *) ((tbpgi.pgi.pg_pde & PG_ADDR) << PG_SHFT32);
  getmem(tbaddr,&tubl,sizeof(tuser_t),"tu-block");
  if( ((proc_t *) tubl.tu_taskp - proca) == (prctbl - proct) )  {
    llong a1,a2;
    a1=tubl.tu_excnt-tubl.tu_vxcnt+tubl.tu_vecnt;
    a2=(llong)( (double)prctbl->p_hutime * (SX3CYCLE*1000.) );
    if(a2 != 0) 
      return((long) (a1/a2));
    else
      return(0);
  } else
    return(-2);
}

long tflops(prctbl) /* return mflops from task block, tmops is prerequisite call */
struct proc *prctbl;
{
  llong a1,a2;
  a1=tubl.tu_fpec;
  a2=(llong)( (double)prctbl->p_hutime * (SX3CYCLE*1000.) );
  if(a2 != 0)
    return((long) (a1/a2));
  else
    return(0);
}

long tvecrat() /* return vectorization ratio */
{
  llong a1,a2;
  a1=tubl.tu_vecnt;
  a2=(tubl.tu_excnt-tubl.tu_vxcnt+tubl.tu_vecnt)/100;
  if(a2 != 0) 
    return((long) (a1/a2));
  else
    return(0); 
}

long tveccpu(prctbl) /* return vector cpu usage ratio */
struct proc *prctbl;
{
  llong a1,a2;
  a1=tubl.tu_vecc;
  a2=prctbl->p_hutime/100;
  if(a2 != 0) 
    return((long) (a1/a2));
  else
    return(0);
}
#endif

displproc(prc,linep,dt)
proc_t *prc;
int    *linep;
mtime_t dt;
{
  char *pgmode="KM";
#ifdef _SX3
  char *pstat=" SRZSIPXW";
  long mops,vcpu;
  llong ttime;
#endif
#ifdef _AIX370
  char *pstat=" SWRIZTX";
#endif
  long tmsec;
  char *lstcmd;
  long ksize;
  int lstcmd_l;
  int line;
  int iprc;

  line = *linep;
#ifdef _SX3
  if(dspmod == 3) {
    mops=tmops(prc);
    vcpu=tveccpu(prc);
    if( tubl.tu_vecc == 0 && vdspsel == 1) {
      *linep = *linep - 1;
      return;
    }
  }
#endif
  iprc=((long)prc-(long)proct)/sizeof(struct proc);
  if(dspmod==0) lstcmd_l=15;
  else lstcmd_l=14;
  lstcmd=lastcmd(prc); /* this reads the U-block header */
  if( strlen(lstcmd) > cmdcolumns ) lstcmd[ cmdcolumns ]='\0';
#ifdef _AIX370
  mvprintw(line,0,"%6d",prc->p_pid);
  mvprintw(line,7,"%-8s",uid2name(prc->p_uid));
#else
  mvprintw(line,0,"%5d",prc->p_pid);
  mvprintw(line,6,"%-8s",uid2name(prc->p_uid));
#endif
  mvprintw(line,14,"%3d",prc->p_nice);
#ifdef _SX3
  mvprintw(line,17,"%3d",prc->p_basepri);
  if(prc->p_jid) { mvprintw(line,21,"%5d",prc->p_jid); } else
    mvprintw(line,21,"     ");
  if(prc->p_pgmode<2) {mvprintw(line,27,"%c",pgmode[prc->p_pgmode]); }
  if(prc->p_task.t_stat <= SRWAIT) {mvprintw(line,29,"%c",pstat[prc->p_task.t_stat]); }
  mvprintw(line,31,"%3d",(prc->p_task).t_pri);
#endif
#ifdef _AIX370
  if(prc->p_stat <= SXBRK) {mvprintw(line,29,"%c",pstat[prc->p_stat]); }
  mvprintw(line,31,"%3d",prc->p_pri);
#endif
#ifdef _DECOSF
  mvprintw(line,24,"%c",syd_res[iprc].statc);
  mvprintw(line,20,"%3d",syd_res[iprc].p_pri);
  mvprintw(line,17,"%3d",syd_res[iprc].p_basepri);
  if(syd_res[iprc].sleep_time != 999) 
    mvprintw(line,25,"%4d",syd_res[iprc].sleep_time);
  else
    mvprintw(line,25," ***");
#endif
#ifdef _SX3
  mvprintw(line,31,"%3d",(prc->p_task).t_pri);
  ksize=(long)((double)prc->p_size * (double) 32.768);
  if(ksize>=1000)
    mvprintw(line,35,"%3dM",ksize/1000);
  else
    mvprintw(line,35,"%3dK",ksize);
  if(dspmod==0) {
    tmsec=(long) ((double)prc->p_hstime * SX3CYCLE);
    mvprintw(line,40,"%6d.%03d",tmsec/1000,tmsec%1000);
    tmsec=(long) ((double)prc->p_hutime * SX3CYCLE);
    mvprintw(line,50,"%6d.%03d",tmsec/1000,tmsec%1000);
    mvprintw(line,60,"%3d%%",(int)((double)(dt*100) / (double)mtotcpu)); /* saved by diff comp */
    mvprintw(line,65,"%-s",lstcmd);
  } else {
    mvprintw(line,39,"%3d%%",(int)((double)(dt*100) / (double)mtotcpu)); /* saved by diff comp */
    ttime=(prc->p_hstime + prc->p_hutime)/100;
    if(ttime != 0)
      mvprintw(line,43,"%3d%%",(long) (prc->p_hutime/ttime));
    else
      mvprintw(line,43,"  0%%");
    if(mops >= 0) {
      mvprintw(line,47,"%5d ",mops);
      mvprintw(line,53,"%4d ",tflops(prc));
      mvprintw(line,57,"%3d%% ",tvecrat());
      mvprintw(line,61,"%3d%% ",vcpu);
    } else
      mvprintw(line,47,"    -    -  -%%  -%% ");
    mvprintw(line,66,"%-s",lstcmd);
  }
#else /* SX3 */
#ifndef _DECOSF
  if (prc->p_stat == SZOMB) {
    mvprintw(line,35,"%3dK",0);
    mvprintw(line,40,"%6d.%03d",0,0);
    mvprintw(line,50,"%6d.%03d",0,0);
    mvprintw(line,60,"%3d%%",0);
    mvprintw(line,65,"%-s","<defunct>");
  } else {
    ksize=(long)((double)syd_res[iprc].size * (double) 4.096);
    if(ksize>=1000)
      mvprintw(line,35,"%3dM",ksize/1000);
    else
      mvprintw(line,35,"%3dK",ksize);
    mvprintw(line,40,"%6d.%03d",syd_res[iprc].stime/1000,syd_res[iprc].stime%1000);
    mvprintw(line,50,"%6d.%03d",syd_res[iprc].utime/1000,syd_res[iprc].utime%1000);
#else /* not _DECOSF */
  if (prc->p_stat == SZOMB) {
    mvprintw(line,35,"%3dK",0);
    mvprintw(line,40,"%6d.%03d",0,0);
    mvprintw(line,50,"%6d.%03d",0,0);
    mvprintw(line,60,"%3d%%",0);
    mvprintw(line,65,"%-s","<defunct>");
  } else {
    ksize= syd_res[iprc].size / 1024;
    if(ksize>=1000)
      if(ksize<10000)
	mvprintw(line,35,"%4d",ksize);
      else
	mvprintw(line,35,"%3dM",ksize/1000);
    else
      mvprintw(line,35,"%3dK",ksize);
    ksize= syd_res[iprc].vsize / 1024;
    if(ksize>=1000)
      if(ksize<10000)
	mvprintw(line,30,"%4d",ksize);
      else
	mvprintw(line,30,"%3dM",ksize/1000);
    else
      mvprintw(line,30,"%3dK",ksize);
    mvprintw(line,40,"%6d.%03d",syd_res[iprc].stime.seconds,syd_res[iprc].stime.microseconds/1000);
    mvprintw(line,50,"%6d.%03d",syd_res[iprc].utime.seconds,syd_res[iprc].utime.microseconds/1000);
#endif /* DECOSF */
    if(mtotcpu==0) mtotcpu=1;
    mvprintw(line,60,"%3d%%",(int)((double)(dt*100) / (double)mtotcpu)); /* saved by diff comp */
    mvprintw(line,65,"%-s",lstcmd);
  }
#endif /* SX3 */
  clrtoeol();
}

displpbar(prc,line,dt) /* bar mode process display */
proc_t *prc;
int    line;
mtime_t dt;
{
  char *lstcmd;
  int perc;
#ifdef _SX3
  if ((prc->p_task).t_stat == SZOMB)
#else
  if (prc->p_stat == SZOMB)
#endif
  {
    perc= 0;
    mvprintw(line,0,"%-8s ",uid2name(prc->p_uid));
    mvprintw(line,9,"%-7s ","<defunc");
    mvprintw(line,16,"%3d%% ",perc); /* saved by diff comp */
  }
  else
  {
    perc=(int)((double)(dt*100) / (double)mtotcpu);
    if (perc < 0) perc=0;
    lstcmd=lastcmd(prc); /* this reads the U-block header */
    lstcmd[7]='\0';
    mvprintw(line,0,"%-8s ",uid2name(prc->p_uid));
    mvprintw(line,9,"%-7s ",lstcmd);
    mvprintw(line,16,"%3d%% ",perc); /* saved by diff comp */
    perc=(int)((double)(dt*BARCHARTMAX) / (double)mtotcpu);
  }
  if((perc<0) || (perc>BARCHARTMAX)) perc=0;
  barc[perc]='\0';
  attron(A_ALTCHARSET);
  mvprintw(line,BARCOL1,"%s",barc);
  attroff(A_ALTCHARSET);
  barc[perc]= barchar;
  clrtoeol();
}

int do_nlist(fn,nl,nlsz)
char *fn;
struct nlist *nl;
int nlsz;
{
  struct stat st1,st2;
  int fd;

  if(stat(SYD_NLIST_FILE,&st1) == 0) {
    if(stat(fn,&st2) != 0) {
      perror("cannot stat kernel"); exit(98);
    } else
      if(st1.st_mtime > st2.st_mtime) { /* our saved nlist is still valid */
	if((fd=open(SYD_NLIST_FILE,O_RDONLY)) < 0) {
	  perror("cannot open nlist file for RDONLY"); exit(98);
	}
	if(read(fd,nl,nlsz) != nlsz) {
	  perror("cannot read nlist file"); exit(98);
	}
	close(fd);
	return;
      }
  }
  unlink(SYD_NLIST_FILE);
  nlist(fn,nl);
  if((fd=open(SYD_NLIST_FILE,O_RDWR|O_CREAT,0600)) < 0) {
    perror("cannot open nlist file RDWR"); exit(98);
  }
  if(write(fd,nl,nlsz) != nlsz) {
    perror("cannot write nlist file"); exit(98);
  }
  close(fd);
  chmod(SYD_NLIST_FILE,0444);
  chown(SYD_NLIST_FILE,0,0);
  return;
}
